// Name:	TclMCi.cpp
// Author:	Farzad Pezeshkpour (fuzz@sys.uea.ac.uk)
// Date:	16th June 1998
// Update:	18th June 1998	General tidyup.
//			05th Aug  2000  Added stub and package support.
//
// Purpose:	Implements multimedia access using Microsoft's 
//			Media Control Interface (MCI). Extensive documentation
//			can be found about MCI on Microsoft site (www.microsoft.com/msdn)
//			Examples files included will show the kind of things that we
//			can do with this, but essentially, TclMci allows developers
//			to i) play audio files from various formats (wav, mp2, mod, mid etc.)
//			  ii) play video files (avi qt mpg etc.) in a separate window
//				  or directly overlayed on a Tk window - again check the demos
//				  for more info.
//			 iii) control peripherals such as Video Disc Recorders 
//			      and VCRs.
//	
//	Enjoy!
//			fuzz.


#include <windows.h>
#include <mmsystem.h>
#include <tcl.h>
#include <string.h>


// DEFINES

#define EXPORT extern "C" _declspec(dllexport)

#define DELETE_ARRAY(x) if (x != NULL) {\
							delete[] x;\
							x = NULL;\
							}

#define DELETE_POINTER(x) if (x != NULL) {\
							delete x;\
							x = NULL;\
							}

// CLASSES

// simple event queue class
class CEventQ 
{
protected:
	typedef struct __Event {
		LONG	dev;
		WPARAM	flags;
		struct __Event *pNext;
	} Event;

	Event *pFirst, *pLast;
	HANDLE			hMutex;

public:
	CEventQ () : pFirst(NULL), pLast(NULL) 
	{
		hMutex = CreateMutex (NULL, FALSE, "EventMutex");
		if (hMutex == NULL) throw "Failed to create mutex";
	};

	virtual ~CEventQ () 
	{
		while (pFirst != NULL)
		{
			pLast = pFirst->pNext;
			delete pFirst;
			pFirst = pLast;
		}
	}


	void Add (LONG devid, WPARAM flags)
	{
		WaitForSingleObject (hMutex, INFINITE);
		Event *pEvent = new Event;
		pEvent->dev = devid;
		pEvent->flags = flags;
		pEvent->pNext = NULL;
		if (pLast != NULL) 
			pLast->pNext = pEvent;
		else 
			pFirst = pEvent;
		pLast = pEvent;
		ReleaseMutex(hMutex);
	}

	bool Remove (LONG &devid, WPARAM &flags)
	{
		WaitForSingleObject (hMutex, INFINITE);
		if (pFirst == NULL) {
			ReleaseMutex (hMutex);	
			return false;
		}
		devid = pFirst->dev;
		flags = pFirst->flags;
		Event *pTemp = pFirst->pNext;
		delete pFirst;
		if ((pFirst = pTemp) == NULL)
			pLast = NULL;
		ReleaseMutex (hMutex);
		return true;
	}

	bool Empty ()
	{
		WaitForSingleObject (hMutex, INFINITE);
		bool empty = (pFirst == NULL);
		ReleaseMutex (hMutex);
		return empty;
	}
};


// command table class
class CCommandTable 
{
protected:
	class Binding
	{
	public:
		Binding (Tcl_Interp *pInterp, Tcl_Obj *device, Tcl_Obj *command) : 
		  m_device (device), m_command(command), m_pInterp (pInterp)
		{
			  Tcl_IncrRefCount (m_device);
			  Tcl_IncrRefCount (m_command);
		}

		~Binding ()
		{
			Tcl_DecrRefCount (m_device);
			Tcl_DecrRefCount (m_command);
		}

		bool Invoke (WPARAM flags)
		{
			Tcl_Obj *pNotify;
			switch (flags)
			{
			case MCI_NOTIFY_ABORTED:
				pNotify = Tcl_NewStringObj ("abort", -1); 
				break;
			case MCI_NOTIFY_FAILURE:
				pNotify = Tcl_NewStringObj ("fail", -1); 
				break;
			case MCI_NOTIFY_SUCCESSFUL:
				pNotify = Tcl_NewStringObj ("ok", -1); 
				break;
			case MCI_NOTIFY_SUPERSEDED:
				pNotify = Tcl_NewStringObj ("supersede", -1); 
				break;
			default:
				pNotify = Tcl_NewStringObj ("unknown", -1); 
				break;
			}
			
			Tcl_Obj *pComObj = Tcl_NewObj();
			Tcl_ListObjAppendElement (m_pInterp, pComObj, m_command);
			Tcl_ListObjAppendElement (m_pInterp, pComObj, m_device);
			Tcl_ListObjAppendElement (m_pInterp, pComObj, pNotify);
			if (Tcl_EvalObj (m_pInterp, pComObj) == TCL_ERROR) {
				Tcl_BackgroundError (m_pInterp);
				return false;
			}
			return true;
		}

	protected:
		Tcl_Interp *m_pInterp;
		Tcl_Obj *m_device;
		Tcl_Obj *m_command;
	};



public:
	CCommandTable ()
	{
		
	}

	virtual ~CCommandTable()
	{
		Tcl_HashSearch search;
		Tcl_HashEntry *pEntry = Tcl_FirstHashEntry (&m_table, &search);
		while (pEntry != NULL)
		{
			Binding *pBinding = (Binding*)Tcl_GetHashValue (pEntry);
			delete pBinding;
			Tcl_DeleteHashEntry(pEntry);
			pEntry = Tcl_NextHashEntry (&search);
		}
		Tcl_DeleteHashTable (&m_table);
	}

	void Init ()
	{
		Tcl_InitHashTable (&m_table, TCL_ONE_WORD_KEYS);
	}

	int SetCommand (Tcl_Interp *pInterp, Tcl_Obj *device, Tcl_Obj* command)
	{
		// Get the device id
		MCIDEVICEID id = mciGetDeviceID (Tcl_GetStringFromObj(device, NULL));
		if (id == 0) {
			Tcl_SetResult (pInterp, "Unable to get device id.", TCL_STATIC);
			return	TCL_ERROR;		
		}

		int created;
		Tcl_HashEntry *pEntry = 
			Tcl_CreateHashEntry (&m_table, (char*)id, &created);
		if (!created) {
			Binding *binding = (Binding*)Tcl_GetHashValue (pEntry);
			delete binding;
		}
		Binding *pBind = new Binding (pInterp, device, command);
		Tcl_SetHashValue (pEntry, (ClientData)pBind);
		return TCL_OK;
	}

	bool Invoke (LONG device, WPARAM flags)
	{
		Tcl_HashEntry *pEntry = Tcl_FindHashEntry (&m_table, (char*)device);
		if (pEntry == NULL) return false;
		Binding *pBind = (Binding*)Tcl_GetHashValue (pEntry);
		return pBind->Invoke (flags);
	}

protected:
	Tcl_HashTable m_table;
};











// Function declarations
//
int		MciSendString (ClientData , Tcl_Interp *pInterp, int objc, Tcl_Obj *CONST objv[]);
int		MciSetCallback(ClientData , Tcl_Interp *pInterp, int objc, Tcl_Obj *CONST objv[]);
LRESULT CALLBACK	MciWndProc (HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
DWORD	WINAPI MciThread (LPVOID hInstanceHandle);
void	MciTclExit (ClientData);
void	MciEventSetup (ClientData cd, int flags);
void	MciEventCheck (ClientData cd, int flags);
int		MciEventProcess(Tcl_Event *pEvent, int flags);




// Globals	
HANDLE			hQuitMutex		= NULL;
HANDLE			hThread			= NULL;
DWORD			dwThreadId		= 0;
HWND			hTclMciWnd		= NULL;
Tcl_Interp		*g_pInterp		= NULL;
char			*szCommand		= NULL;
CEventQ			events;
CCommandTable	commands;




/* Function Definitions *\
\* ==================== */


// RegisterClass
// Purpose: Declares the WNDCLASS of the Tcl mci window
//			and registers it.
//
void RegisterClass (HINSTANCE hInstance)
{
	WNDCLASS wc;
	wc.lpszClassName = "FuzTclMciClass";
	wc.lpfnWndProc = (WNDPROC) MciWndProc;
	wc.style = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon( NULL, IDI_APPLICATION );
	wc.hCursor = LoadCursor( NULL, IDC_ARROW );
	wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH); 
	wc.lpszMenuName = "MainMenu";
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	if (RegisterClass( &wc ) == 0) {
		LPVOID lpMsgBuf;
		FormatMessage( 
			FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL, GetLastError(), 
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			(LPTSTR) &lpMsgBuf,
			0,
			NULL );

		MessageBox( NULL, (char*)lpMsgBuf, "GetLastError", MB_OK|MB_ICONINFORMATION );
		LocalFree( lpMsgBuf );
	}
}





// MciWndProc
// Purpose:	Handles messages from the TclMci window
//			Queues notifications from mci for processing by 
//			the registered Tcl event handler
//
LRESULT CALLBACK MciWndProc (
	HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam
) 
{
	int dummy = 0;
	switch (uMsg) {
	case MM_MCINOTIFY:
		// get the device id and the flags
		events.Add (lParam, wParam);
		break;
	case WM_DESTROY:
		return 0;
	default:
		break;
	}
	return DefWindowProc( hwnd, uMsg, wParam, lParam);
}



// Dll entry point
// Purpose: Create mutexes and the mci window thread
//			Returns true iff all operations successful.
//
BOOL WINAPI DllMain ( HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	switch (fdwReason) {
	case DLL_PROCESS_ATTACH:
		hQuitMutex = CreateMutex (NULL, TRUE, "TclMciQuitMutex");
		if (hQuitMutex == NULL) {
			return FALSE;
		}

		hThread = CreateThread (NULL, 0, MciThread, (LPVOID)hinstDLL, NULL, &dwThreadId);
		if (hThread == NULL) {
			CloseHandle (hQuitMutex);
			hQuitMutex = NULL;
			return FALSE;
		}
		break;
	case DLL_PROCESS_DETACH:
		break;
	case DLL_THREAD_ATTACH:
		break;
	case DLL_THREAD_DETACH:
		break;
	}
	return TRUE;
}
 

// Mci_Init
// Purpose:	Secondary initialisation of DLL. Registers commands and
//			the exit handler.
//
EXPORT int Mci_Init (Tcl_Interp *pint) 
{
#ifdef USE_TCL_STUBS
#if (TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION >= 1)
		// initialise the Tcl stubs - failure is very bad
		if (Tcl_InitStubs (pint, "8.1", 0) == NULL)
			return TCL_ERROR;
/*
		// if Tk is loaded then initialise the Tk stubs
		if (Tcl_Eval (pInterp, "package present Tk") != TCL_ERROR) {
			// initialise the Tk stubs - failure 
			if (Tk_InitStubs (pInterp, "8.0", 0) == NULL)
				return TCL_ERROR;
			g_bTkInit = true;
		}

*/
		
#else
#error Wrong Tcl version for Stubs
#endif // (TCL_MAJOR_VERSION == 8) && (TCL_MINOR_VERSION >= 1)
#endif // USE_TCL_STUBS

	Tcl_PkgProvide (pint, "tmci", "1.1");
	commands.Init();
	g_pInterp = pint;
	Tcl_CreateObjCommand (g_pInterp, "mci", MciSendString, NULL, NULL);
	Tcl_CreateObjCommand (g_pInterp, "mci_notify", MciSetCallback, NULL, NULL);
	Tcl_CreateExitHandler (MciTclExit, NULL);
	Tcl_CreateEventSource (MciEventSetup, MciEventCheck, NULL);
	return TCL_OK;
}




// MciSendString
// Purpose:	Implements the 'mci ...' command. A simple and complete wrapper
//			for the mciSendString command.
//
int MciSendString (ClientData cd, Tcl_Interp *pInterp, int objc, Tcl_Obj *CONST objv[])
{
#define size 2048
	char ret[size];
	MCIERROR er;

	if (objc != 2 ) {
		Tcl_SetResult (pInterp, "mci requires a command.", TCL_STATIC);
		return TCL_ERROR;
	}

	er = mciSendString (Tcl_GetStringFromObj (objv[1], NULL), ret, size, hTclMciWnd);
	if (er == 0) {
		Tcl_SetResult (pInterp, ret, TCL_VOLATILE);
		return TCL_OK;
	}

	if (mciGetErrorString (er, ret, size))
		Tcl_SetResult (pInterp, ret, TCL_VOLATILE);
	else
		Tcl_SetResult (pInterp, "Unknown error.", TCL_STATIC);

	return TCL_ERROR;
}


// MciThread
// Purpose:	Mci window thread
//			Registers and creates the Tcl mci window
//			The message pump incorporates a check on
//			the quit mutex being signalled.
//			Note that WM_DESTROY never really arrives at the WindowProc
//			of the window. Shouldn't create any problems, but let
//			me know if so (fuzz@sys.uea.ac.uk).
//
DWORD WINAPI MciThread (LPVOID hInstanceHandle)
{

	RegisterClass ((HINSTANCE)hInstanceHandle);
	hTclMciWnd = CreateWindow ("FuzTclMciClass", "FuzTclMci", 
		NULL , 0, 0, 
		CW_USEDEFAULT, CW_USEDEFAULT, (HWND)NULL, (HMENU)NULL, 
		(HINSTANCE)hInstanceHandle, (LPVOID)NULL);

	if (hTclMciWnd == NULL) return 0;

	while (TRUE) {
		DWORD ret = MsgWaitForMultipleObjects (1, &hQuitMutex, FALSE, INFINITE, QS_ALLINPUT);
		if (ret == WAIT_OBJECT_0)
			return 0;
		
		MSG msg ; 
		while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) { 
			if (msg.message == WM_QUIT)
				return 0; 
			DispatchMessage(&msg); 
		}
	}
}



// MciTclExit
// Purpose: To cleanly shutdown the window thread
// Method:	Signals the thread and waits on it (for 2 seconds)
//			Kills thread if not successful.
//			Closes off the handles to the mutexes
//
void MciTclExit (ClientData)
{
	if (hThread != NULL) {
		ReleaseMutex (hQuitMutex);
		if (WaitForSingleObject (hThread, 2000) != WAIT_OBJECT_0) {
			TerminateThread (hThread, 0);
		}
		CloseHandle(hThread);
		hThread = NULL;
	}

	CloseHandle (hQuitMutex);
	hQuitMutex = NULL;
}


// format of mci_setCallback
//	mci_notify device command
int MciSetCallback(ClientData , Tcl_Interp *pInterp, int objc, Tcl_Obj *CONST objv[])
{
	if (objc != 3) {
		Tcl_SetResult (pInterp, "command format: mci_notify <device> <command>", TCL_STATIC);
		return	TCL_ERROR;
	}

	return commands.SetCommand (pInterp, objv[1], objv[2]);
}





void MciEventSetup (ClientData cd, int flags)
{

}


void MciEventCheck (ClientData cd, int flg)
{
	if (events.Empty()) return;
	Tcl_Event *pEvent = (Tcl_Event*)Tcl_Alloc (sizeof(Tcl_Event));
	pEvent->proc = MciEventProcess;
	Tcl_QueueEvent (pEvent, TCL_QUEUE_TAIL);
}



int MciEventProcess (Tcl_Event *pEvent, int flg)
{
	LONG dev;
	WPARAM flags;
	if (!events.Remove (dev, flags)) return 1;
	commands.Invoke (dev, flags);
	return 1;
}